unit cDIBCompressor;

{-----------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/MPL-1.1.html

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: cDIBCompress.PAS, released September 04, 2000.

The Initial Developer of the Original Code is Peter Morris (pete@stuckindoors.com),
Portions created by Peter Morris are Copyright (C) 2000 Peter Morris.
All Rights Reserved.

Purpose of file:
To handle the compression / decompression of DIB data, and to allow people to register
their own compressors.

Contributor(s):
None as yet


Last Modified: September 04, 2000
Current Version: 1.00

You may retrieve the latest version of this file at the Project JEDI home page,
located at http://www.delphi-jedi.org
or at http://www.stuckindoors.com/dib


Known Issues:
-----------------------------------------------------------------------------}


interface
uses
  Classes, SysUtils, Windows;

type
  TAbstractDIBCompressor = class(TPersistent)
  private
  protected
  public
    function  GetGUID : TGUID; virtual; abstract;
    function  CanDecompress(const GUID : TGUID) : Boolean; virtual; abstract;
    function  Compress(const Source, Dest; const SourceSize : DWord) : DWord; virtual; abstract;
    procedure Decompress(const GUID : TGUID; const Source, Dest; const SourceSize, DestSize : DWord); virtual; abstract;
    function  GetDisplayName : String; virtual; abstract;

    function  GetAboutText : String; virtual;
    function  GetAuthor : String; virtual;
    function  GetEmail : String; virtual;
    function  GetURL : String; virtual;
  end;

procedure RegisterDIBCompressor(const Compressor : TAbstractDIBCompressor);
function  CompressorCount : Integer;
function  Compressor(Index : Integer) : TAbstractDIBCompressor;
procedure Compress(const Source, Dest : TStream);
procedure Decompress(const Source, Dest : TStream);
function  FindCompressor(const GUID : String) : TAbstractDIBCompressor;

var
  DefaultCompressor : TAbstractDIBCompressor;

implementation
uses
  COMObj;

type
  EDIBCompressError = class(Exception);

const
  cCompressedSig = 'DIBCMP';

var
  FList : TList;

procedure RegisterDIBCompressor(const Compressor : TAbstractDIBCompressor);
begin
  FList.Add(Compressor);
end;

function  CompressorCount : Integer;
begin
  Result := FList.Count;
end;

function  Compressor(Index : Integer) : TAbstractDIBCompressor;
begin
  if (Index < 0) or (Index >= FList.Count) then
    raise EDIBCompressError.Create('Index ' + IntToStr(Index) + ' out of range.');
  Result := TAbstractDIBCompressor(FList[Index]);
end;

procedure Compress(const Source, Dest : TStream);
var
  MSSource,
  MSDest                      : TMemoryStream;
  GUID                        : TGUID;
  OrigPosition                : Integer;
  NewDataSize,
  DataSize                    : DWord;
  pSource                     : Pointer;
begin
  if Source is TMemoryStream then
    MSSource := TMemoryStream(Source)
  else begin
    MSSource := TMemoryStream.Create;
    MSSource.CopyFrom(Source, Source.Size);
    MSSource.Seek(0,0);
  end;

  MSDest := TMemoryStream.Create;
  try
    if not Assigned(DefaultCompressor) then
      Dest.CopyFrom(Source, Source.Size)
    else begin
      OrigPosition := Source.Position;
      MSDest.SetSize(MSSource.Size - MSSource.Position);
      pSource := Pointer(Integer(MSSource.Memory) + MSSource.Position);
      NewDataSize :=
        DefaultCompressor.Compress(pSource^,MSDest.Memory^, MSSource.Size - MSSource.Position);
      if NewDataSize >0 then begin
        MSDest.SetSize(NewDataSize);
        GUID := DefaultCompressor.GetGuid;
        Dest.Write(cCompressedSig, Length(cCompressedSig));
        Dest.Write(GUID, SizeOf(GUID));
        DataSize := Source.Size;
        Dest.Write(DataSize, SizeOf(DWord));
        MSDest.Seek(0, soFromBeginning);
        Dest.CopyFrom(MSDest, MSDest.Size);
      end else begin
        Source.Seek(OrigPosition, soFromBeginning);
        Dest.CopyFrom(Source, Source.Size);
      end;
    end;
  finally
    MSDest.Free;
    if MSSource <> Source then MSSource.Free;
  end;
end;

procedure Decompress(const Source, Dest : TStream);
var
  I                           : Integer;
  FCompressor                 : TAbstractDIBCompressor;
  Signature                   : Array[0..5] of char;
  GUID                        : TGUID;
  NeedCompressor              : Boolean;
  OrigPosition                : Integer;
  DataSize                    : DWord;
  MSSource,
  MSDest                      : TMemoryStream;
  pSource                      : Pointer;
begin
  NeedCompressor := False;
  OrigPosition := Source.Position;
  DataSize := 0;
  if Source.Size >= Length(cCompressedSIG) + SizeOf(TGUID) + SizeOf(DWord) then begin
    Source.Read(Signature[0],6);
    if Signature = cCompressedSig then begin
      Source.Read(GUID, SizeOf(GUID));
      Source.Read(DataSize, SizeOf(DWord));
      NeedCompressor := True;
    end else
      Source.Seek(OrigPosition,soFromBeginning);
  end;

  if not NeedCompressor then
    Dest.CopyFrom(Source, Source.Size)
  else begin
    FCompressor := nil;
    for I:=FList.Count-1 downto 0 do
      with TAbstractDIBCompressor(FList[I]) do
        if CanDecompress(GUID) then begin
          FCompressor := TAbstractDIBCompressor(FList[I]);
          break;
        end;

    if FCompressor = nil then
      raise EDIBCompressError.Create('Could not find a suitable decompressor.')
    else begin
      //If a memory stream, point to that data
      if Source is TMemoryStream then
        MSSource := TMemoryStream(Source)
      else begin
        //If not a memory stream, Copy the data to a memory stream
        MSSource := TMemoryStream.Create;
        MSSource.CopyFrom(Source, Source.Size - Source.Position);
        MSSource.Seek(0, soFromBeginning);
      end;

      pSource := Pointer(Integer(MSSource.Memory) + MSSource.Position);

      MSDest := TMemoryStream.Create;
      try
        MSDest.SetSize(DataSize);
        FCompressor.Decompress(GUID, pSource^ ,MSDest.Memory^, MSSource.Size - MSSource.Position, DataSize);
        MSDest.Seek(0, soFromBeginning);
        Dest.CopyFrom(MSDest,MSDest.Size);
      finally
        MSDest.Free;
        if MSSource <> Source then MSSource.Free;
      end;
    end;
  end;
end;



{ TAbstractDIBCompressor }

function TAbstractDIBCompressor.GetAboutText: String;
begin
  Result := 'No information supplied';
end;

function TAbstractDIBCompressor.GetAuthor: String;
begin
  Result := 'No information supplied';
end;

function TAbstractDIBCompressor.GetEmail: String;
begin
  Result := '';
end;

function TAbstractDIBCompressor.GetURL: String;
begin
  Result := '';
end;

function FindCompressor(const GUID : String) : TAbstractDIBCompressor;
var
  I: Integer;
begin
  Result := nil;
  for I:=0 to FList.Count-1 do
    if GUIDToString(TAbstractDIBCompressor(FList[I]).GetGUID) = GUID then begin
      Result := TAbstractDIBCompressor(FList[I]);
      break;
    end;
end;

initialization
  FList := TList.Create;
finalization
  while FList.Count > 0 do begin
    TAbstractDIBCompressor(FList[0]).Free;
    FList.Delete(0);
  end;

end.
